
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <libusb.h>
#include <sys/time.h>

#define EP_DATA			(2 | LIBUSB_ENDPOINT_OUT)
#define DATA_LENGTH		1024


#define NUM_SAMPLES_BUFF	(DATA_LENGTH*1024)

static struct libusb_device_handle *devh = NULL;
static float in_samples_buf[NUM_SAMPLES_BUFF];
static short out_samples_buf[2][NUM_SAMPLES_BUFF];
static struct libusb_transfer *samples_transfer0 = NULL;
static struct libusb_transfer *samples_transfer1 = NULL;
static int do_exit = 0;
static int input_is_ready0 = 0;
static int input_is_ready1 = 0;
static int current_transfer = 0;
static FILE *input_fh = NULL;


static pthread_t poll_thread;
static pthread_cond_t exit_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t exit_cond_lock = PTHREAD_MUTEX_INITIALIZER;

static pthread_t input_thread;
static pthread_cond_t input_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t input_cond_lock = PTHREAD_MUTEX_INITIALIZER;

static void request_exit(int code)
{
	do_exit = code;
	pthread_cond_signal(&exit_cond);
}

static void *poll_thread_main(void *arg)
{
	int r = 0;
	printf("poll thread running\n");

	while (!do_exit) {
		struct timeval tv = { 1, 0 };
		r = libusb_handle_events_timeout(NULL, &tv);
		if (r < 0) {
			request_exit(2);
			break;
		}
	}

	printf("poll thread shutting down\n");
	return NULL;
}

//! Byte swap short
int16_t swap_int16( int16_t val ) 
{
    return (val << 8) | ((val >> 8) & 0xFF);
}

static void *input_thread_main(void *arg)
{
	int i;
	printf("reader thread running\n");

	while (!do_exit) {

		input_is_ready0 = 0;
		input_is_ready1 = 0;
		if(fread(in_samples_buf, NUM_SAMPLES_BUFF, sizeof(float), input_fh) < 1) {
			request_exit(2);
			break;
		}

		if(current_transfer==0) {
			short *out_buf = out_samples_buf[0];
			for(i=0;i<NUM_SAMPLES_BUFF;i++) {
//				out_buf[i] = in_samples_buf[i]; //[-32767.0 ... 32767.0] => [-32767 ... 32767]
				out_buf[i] = (short) in_samples_buf[i]; //[-32767.0 ... 32767.0] => [-32767 ... 32767]
				out_buf[i] = swap_int16(out_buf[i]);
			}
			input_is_ready0 = 1;
		} else {
			short *out_buf = out_samples_buf[1];
			for(i=0;i<NUM_SAMPLES_BUFF;i++) {
				out_buf[i] = (short) in_samples_buf[i]; //[-32767.0 ... 32767.0] => [-32767 ... 32767]
//				out_buf[i] = (in_samples_buf[i] & 0x00ff) << 8 | ((in_samples_buf[i] >> 8) & 0x00ff); //[-32767.0 ... 32767.0] => [-32767 ... 32767]
//				out_buf[i] = in_samples_buf[i]; //[-32767.0 ... 32767.0] => [-32767 ... 32767]
				//! Byte swap short
				out_buf[i] = swap_int16(out_buf[i]);
			}
			input_is_ready1 = 1;
		}

		pthread_mutex_lock(&input_cond_lock);
		pthread_cond_wait(&input_cond, &input_cond_lock);
		pthread_mutex_unlock(&input_cond_lock);
	}

	printf("reader thread shutting down\n");
	return NULL;
}

static int find_sdrtx_device(void)
{
	devh = libusb_open_device_with_vid_pid(NULL, 0xfffe, 0x0002);
	return devh ? 0 : -EIO;
}


struct timespec start_time = {0,0}, end_time;
double elapsed;
static int total_samples= 0;

static int show_info(short *data)
{
	//printf("readed to mem\n");

	//fprintf(stderr, ".");
	total_samples += NUM_SAMPLES_BUFF;

	if((current_transfer==0 && input_is_ready0 == 1) ||
	   (current_transfer==1 && input_is_ready1 == 1))
		fprintf(stderr, ".");
	else
		fprintf(stderr, "x");

	//memcpy(samples_buf, input_buf, sizeof(samples_buf));
	clock_gettime(CLOCK_REALTIME,  &end_time);
	elapsed = (end_time.tv_sec - start_time.tv_sec) + (double) (end_time.tv_nsec - start_time.tv_nsec) * 1e-9;
	if(elapsed >= 1) {
		fprintf(stderr, ": samples = %.3f kS/s, elapsed=%.3f s\n",
			(double) total_samples/(elapsed * 1000), elapsed);
		start_time = end_time;
		total_samples=0;
	}

	return 0;
}

static void LIBUSB_CALL cb_img0(struct libusb_transfer *transfer)
{
	if (transfer->status != LIBUSB_TRANSFER_COMPLETED) {
		fprintf(stderr, "samples transfer status %d?\n", transfer->status);
		samples_transfer0 = NULL;
		request_exit(2);
		return;
	}

	show_info(out_samples_buf[0]);
	current_transfer=0;
	pthread_cond_signal(&input_cond);

	if (libusb_submit_transfer(samples_transfer1) < 0)
		request_exit(2);
}

static void LIBUSB_CALL cb_img1(struct libusb_transfer *transfer)
{
	if (transfer->status != LIBUSB_TRANSFER_COMPLETED) {
		fprintf(stderr, "samples transfer status %d?\n", transfer->status);
		samples_transfer1 = NULL;
		request_exit(2);
		return;
	}

	show_info(out_samples_buf[1]);
	current_transfer=1;
	pthread_cond_signal(&input_cond);

	if (libusb_submit_transfer(samples_transfer0) < 0)
		request_exit(2);

}

static int init_player(void)
{
	int r;

	r = libusb_submit_transfer(samples_transfer0);
	if (r < 0)
		return r;

	return r;
}

static int alloc_transfers(void)
{
	samples_transfer0 = libusb_alloc_transfer(0);
	if (!samples_transfer0)
		return -ENOMEM;

	samples_transfer1 = libusb_alloc_transfer(0);
	if (!samples_transfer1)
		return -ENOMEM;

	libusb_fill_bulk_transfer(samples_transfer0, devh, EP_DATA, (unsigned char *) out_samples_buf[0],
		NUM_SAMPLES_BUFF*sizeof(short), cb_img0, NULL, 0);

	libusb_fill_bulk_transfer(samples_transfer1, devh, EP_DATA, (unsigned char *) out_samples_buf[1],
		NUM_SAMPLES_BUFF*sizeof(short), cb_img1, NULL, 0);

	return 0;
}

static void sighandler(int signum)
{
	request_exit(1);
}

int main(int argc, char **argv)
{
	struct sigaction sigact;
	int r = 1;

#if 0
	input_fh = fopen(argv[1], "r");
	fprintf(stderr, "input: %s\n", argv[1]);
	int xx=0;
	while(fread(in_samples_buf, 2, sizeof(float), input_fh) > 0) {
		fprintf(stderr, "[%.1f %.1f] ", in_samples_buf[0], in_samples_buf[1]);
		if(xx%8==0) fprintf(stderr, "\n");
		if( in_samples_buf[0] > 32768.0f ||
		    in_samples_buf[0] < -32768.0f ||
		    in_samples_buf[1] > 32768.0f ||
		    in_samples_buf[1] < -32768.0f)
		    fprintf(stderr, "\n BAD SAMPLE: [%.1f %.1f]\n", in_samples_buf[0], in_samples_buf[1]);
		xx++;
	}
	fprintf(stderr, "\n");
#endif

	r = libusb_init(NULL);
	if (r < 0) {
		fprintf(stderr, "failed to initialise libusb\n");
		exit(1);
	}

	r = find_sdrtx_device();
	if (r < 0) {
		fprintf(stderr, "Could not find/open device\n");
		goto out;
	}

	r = libusb_claim_interface(devh, 0);
	if (r < 0) {
		fprintf(stderr, "usb_claim_interface error %d %s\n", r, strerror(-r));
		goto out;
	}
	printf("claimed interface\n");

	/* async from here onwards */

	sigact.sa_handler = sighandler;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = 0;
	sigaction(SIGINT, &sigact, NULL);
	sigaction(SIGTERM, &sigact, NULL);
	sigaction(SIGQUIT, &sigact, NULL);

	r = pthread_create(&poll_thread, NULL, poll_thread_main, NULL);
	if (r)
		goto out_deinit;

	if(argc >= 2) {
		input_fh = fopen(argv[1], "r");
		fprintf(stderr, "input: %s\n", argv[1]);
	} else
		input_fh = stdin;

	if(!input_fh) {
		fprintf(stderr, "cant open input file\n");
		goto out_deinit;
	}

	r = pthread_create(&input_thread, NULL, input_thread_main, NULL);
	if (r)
		goto out_deinit;

	r = alloc_transfers();
	if (r < 0) {
		request_exit(1);
		pthread_join(poll_thread, NULL);
		goto out_deinit;
	}

	r = init_player();
	if (r < 0) {
		request_exit(1);
		pthread_join(poll_thread, NULL);
		goto out_deinit;
	}

	while (!do_exit) {
		pthread_mutex_lock(&exit_cond_lock);
		pthread_cond_wait(&exit_cond, &exit_cond_lock);
		pthread_mutex_unlock(&exit_cond_lock);
	}

	printf("shutting down...\n");
	pthread_join(poll_thread, NULL);

	if(current_transfer == 0) {
		r = libusb_cancel_transfer(samples_transfer0);
		if (r < 0) {
			request_exit(1);
			goto out_deinit;
		}
	} else {
		r = libusb_cancel_transfer(samples_transfer1);
		if (r < 0) {
			request_exit(1);
			goto out_deinit;
		}
	}


	if(current_transfer == 0) {
		while (samples_transfer0)
			if (libusb_handle_events(NULL) < 0)
				break;
	} else {
		while (samples_transfer1)
			if (libusb_handle_events(NULL) < 0)
				break;
	}

	if (do_exit == 1)
		r = 0;
	else
		r = 1;

	pthread_cond_signal(&input_cond);
	pthread_join(input_thread, NULL);


out_deinit:
	libusb_free_transfer(samples_transfer0);
	libusb_free_transfer(samples_transfer1);
	libusb_release_interface(devh, 0);
out:
	libusb_close(devh);
	libusb_exit(NULL);
	return r >= 0 ? r : -r;
}
