/*-----------------------------------------------------------------------------
 * Code that turns a Cypress FX2 USB Controller into an USB JTAG adapter
 *-----------------------------------------------------------------------------
 * Copyright (C) 2005..2007 Kolja Waschk, ixo.de
 *-----------------------------------------------------------------------------
 * Check hardware.h/.c if it matches your hardware configuration (e.g. pinout).
 * Changes regarding USB identification should be made in product.inc!
 *-----------------------------------------------------------------------------
 * This code is part of usbjtag. usbjtag is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. usbjtag is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  You should have received a
 * copy of the GNU General Public License along with this program in the file
 * COPYING; if not, write to the Free Software Foundation, Inc., 51 Franklin
 * St, Fifth Floor, Boston, MA  02110-1301  USA
 *-----------------------------------------------------------------------------
 */

#include "isr.h"
#include "timer.h"
#include "delay.h"
#include "fx2regs.h"
#include "fx2utils.h"
#include "usb_common.h"
#include "usb_descriptors.h"
#include "usb_requests.h"

#include "syncdelay.h"
#include "i2c.h"
#include "spi.h"

//-----------------------------------------------------------------------------
// Global data

typedef bit BOOL;
#define FALSE 0
#define TRUE  1

xdata unsigned char i2cbuf[4];
xdata unsigned char spibuf[8];

//-----------------------------------------------------------------------------

void usb_fifo_init(void)              // Called once at startup
{
   WORD tmp;

   spi_init();

   CPUCS = bmCLKOE | bmCLKSPD1; SYNCDELAY; // CLKOE=outEnabled, CLK=48Mhz
   IFCONFIG = bmIFCLKSRC | bm3048MHZ | bmASYNC | bmIFCFG1 | bmIFCFG0; SYNCDELAY; //IFCLKSRC=internal, 3048MHZ=48MHz, IFCFG=SlaveFIFO

   FIFORESET  = 0x80; SYNCDELAY;
   FIFORESET  = 0x82; SYNCDELAY;
   FIFORESET  = 0x84; SYNCDELAY;
   FIFORESET  = 0x86; SYNCDELAY;
   FIFORESET  = 0x88; SYNCDELAY;
   FIFORESET  = 0x00; SYNCDELAY;

//   PINFLAGSAB = 0x98; SYNCDELAY; // TODO
//   PINFLAGSCD = 0xfe; SYNCDELAY; // TODO
//   PORTACFG &= ~0xc0; SYNCDELAY; // TODO
   PINFLAGSAB = 0x00; SYNCDELAY;
   PINFLAGSCD = 0x00; SYNCDELAY;
   PORTACFG   = 0x00; SYNCDELAY; //We can use PA.0/INT0, PA.1/INT1, PA.3/WU2, PA.7/FLAGD/SLCS. PA0+PA1+PA3 = SPI. PA7 - data enabler
   FIFOPINPOLAR = 0x00; SYNCDELAY; //active-low

   // Make Timer2 reload at 100 Hz to trigger Keepalive packets

   tmp = 65536 - ( 48000000 / 12 / 100 );
   RCAP2H = tmp >> 8;
   RCAP2L = tmp & 0xFF;
   CKCON = 0; // Default Clock
   T2CON = 0x04; // Auto-reload mode using internal clock, no baud clock.

   // define endpoint configuration

//   REVCTL = 0; SYNCDELAY;
   REVCTL = 0x03; SYNCDELAY;

   EP1OUTCFG  = 0x20; SYNCDELAY;
   EP1INCFG   = 0x20; SYNCDELAY;
//   EP2CFG     = 0xA0; SYNCDELAY; // EP2 is DIR=OUT, TYPE=BULK, SIZE=512, BUF=4x
//   EP2CFG     = 0xA2; SYNCDELAY; // EP2 is DIR=OUT, TYPE=BULK, SIZE=512, BUF=2x
   EP2CFG     = 0xA8; SYNCDELAY; // EP2 is DIR=OUT, TYPE=BULK, SIZE=1024, BUF=4x
   EP4CFG     = 0x20; SYNCDELAY; // EP4 not used
   EP6CFG     = 0x60; SYNCDELAY; // EP6 not used
   EP8CFG     = 0x60; SYNCDELAY; // EP8 not used

   EP2FIFOCFG = 0x00; SYNCDELAY; //INFM1=0, OEP1=0, AUTOOUT=0, AUTOIN=0, ZEROLENIN=0, WORDWIDE=0
   SYNCDELAY;

   EP2BCL = 0x00; SYNCDELAY; //EP2,6 can be 512 or 1024. Byte count low
   EP2BCL = 0x00; SYNCDELAY;

   OUTPKTEND = 0x02; SYNCDELAY; // EP2 is SKIP=no
   OUTPKTEND = 0x02; SYNCDELAY; // EP2 is SKIP=no

   EP2FIFOCFG = 0x11; SYNCDELAY; //INFM1=0, OEP1=0, AUTOOUT=1, AUTOIN=0, ZEROLENIN=0, WORDWIDE=1

   PORTACFG=0x00; SYNCDELAY; //We can use PA.0/INT0, PA.1/INT1, PA.3/WU2, PA.7/FLAGD/SLCS. PA0+PA1+PA3 = SPI. PA7 - data enabler

//   FIFOPINPOLAR=0x00; SYNCDELAY; //default polarities: SLRD active low
   FIFOPINPOLAR=0x3F; SYNCDELAY; //default polarities: SLRD active low
//   EP6FIFOCFG = 0x0c; SYNCDELAY;

//   EP6AUTOINLENH = 2;
//   EP6AUTOINLENL = 0;
}

//-----------------------------------------------------------------------------
// Handler for Vendor Requests (
//-----------------------------------------------------------------------------

unsigned char app_vendor_cmd(void)
{
  // OUT requests. Pretend we handle them all...

  if ((bRequestType & bmRT_DIR_MASK) == bmRT_DIR_OUT)
  {
    if(bRequest == 0x01) {
      i2cbuf[0] = wValueH;
      i2cbuf[1] = wValueL;
      i2cbuf[2] = wIndexH;
      i2cbuf[3] = wIndexL;
      i2c_write(0x61, i2cbuf, 4);
    } else if(bRequest == 0x04) {
      spibuf[4] = wValueH;
      spibuf[5] = wValueL;
      spibuf[6] = wIndexH;
      spibuf[7] = wIndexL;
    } else if(bRequest == 0x03) {
      spibuf[0] = wValueH;
      spibuf[1] = wValueL;
      spibuf[2] = wIndexH;
      spibuf[3] = wIndexL;
    } else if(bRequest == 0x02) {
      spi_xfer(wIndexL, (wIndexH<8) ? wIndexH : 8, spibuf);
    }
    return 1;
  }

  // IN requests.

  if(bRequest == 0x01) {
    i2c_read(0x61, EP0BUF, 1);

    EP0BCH = 0;
    EP0BCL = (wLengthL<1) ? wLengthL : 1;
  } else if(bRequest == 0x02) {
    EP0BUF[0] = spibuf[0];
    EP0BUF[1] = spibuf[1];
    EP0BUF[2] = spibuf[2];
    EP0BUF[3] = spibuf[3];
    EP0BUF[4] = spibuf[4];
    EP0BUF[5] = spibuf[5];
    EP0BUF[6] = spibuf[6];
    EP0BUF[7] = spibuf[7];

    EP0BCH = 0;
    EP0BCL = (wLengthL<8) ? wLengthL : 8;
  }

  return 1;
}

//-----------------------------------------------------------------------------

static void main_loop(void)
{
  while(1)
  {
    if(usb_setup_packet_avail()) usb_handle_setup_packet();
  }
}

//-----------------------------------------------------------------------------

void main(void)
{
  EA = 0; // disable all interrupts

  usb_fifo_init();
  setup_autovectors ();
  usb_install_handlers ();


  EA = 1; // enable interrupts

  fx2_renumerate(); // simulates disconnect / reconnect

  main_loop();
}
